
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"

#ifndef	CSI_BS
#define CSI_BS
#endif

#include "csi.h"

extern void	Error();

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

static Cell	***Read(pbs,Bx,By,Bz,Mx,My,Mz)
PBS		*pbs;
int		Bx,By,Bz,Mx,My,Mz;
{
	Cell	***cell,**temp;
	char	*mae="memory allocation error.";
	int	z,y,Nx,Ny,x;

	if ((cell=ALLOC(Cell **,Mz-Bz))==NULL) Error(mae);

	cell-=Bz;

	for (z=Bz; z<Mz; z++) {
	    if ((cell[z]=ALLOC(Cell *,My-By))==NULL) Error(mae);

	    cell[z]-=By;

	    for (y=By; y<My; y++) {
		if ((cell[z][y]=ALLOC(Cell,Mx-Bx))==NULL) Error(mae);

		cell[z][y]-=Bx;
	    }
	    ReadImageFile(pbs[z].path,&Nx,&Ny,NULL,&temp,NULL);

	    for (y=Ny-1; y>=My; y--) free(temp[y]);

	    for (y=My-1; y>=Ny; y--)
		for (x=Mx-1; x>=Bx; x--) cell[z][y][x]=0;

	    for (; y>=By; y--) {
		for (x=Mx-1; x>=Nx; x--) cell[z][y][x]=0;

		for (; x>=Bx; x--) cell[z][y][x]=temp[y][x];

		free(temp[y]);
	    }
	    for (; y>=0; y--) free(temp[y]);

	    free(temp);
	}
	return cell;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	PBS		*PBSg,*PBSh,*BSg,*BSh;
	int		Ngx,Ngy,Ngz,Nhx,Nhy,Nhz,Dx1,Dy1,Dz1,Dx2,Dy2,Dz2,
			Nghx,Nghy,Nghz,dz,zg,zh,Nz,dy,yg,yh,Ny,dx,xg,xh,Nx,
			z,y,x;
	Cell		***G,***H,***Gz,***Hz,**Gy,**Hy,*Gx,*Hx,g,h;
	double		sqrs,Bgh,Sg,Sh,diff;
	unsigned long	pels;

	if (argc!=5 && argc!=11)
	Error("usage : rmsd_3d dir1 NBS1 dir2 NBS2 {Dx1 Dy1 Dz1 Dx2 Dy2 Dz2}");

	PBSg=CheckSliceImages(argv[1],argv[2],&Ngx,&Ngy,&Ngz,NULL);
	PBSh=CheckSliceImages(argv[3],argv[4],&Nhx,&Nhy,&Nhz,NULL);
	
	if (argc==5) {
	    Dx1=1-Nhx; Dx2=Ngx-1;
	    Dy1=1-Nhy; Dy2=Ngy-1;
	    Dz1=1-Nhz; Dz2=Ngz-1;
	}
	else
	    if ((Dx1=atoi(argv[ 5]))>
		(Dx2=atoi(argv[ 8])) || Dx1<=(-Nhx) || Dx2>=Ngx ||
		(Dy1=atoi(argv[ 6]))>
		(Dy2=atoi(argv[ 9])) || Dy1<=(-Nhy) || Dy2>=Ngy ||
		(Dz1=atoi(argv[ 7]))>
		(Dz2=atoi(argv[10])) || Dz1<=(-Nhz) || Dz2>=Ngz)
		Error("bad range of distance.");

	Nghx=Ngx-Nhx;
	Nghy=Ngy-Nhy;
	Nghz=Ngz-Nhz;

	G=Read(PBSg,(Dx1<0)?0: Dx1,
		    (Dy1<0)?0: Dy1,
		    (Dz1<0)?0: Dz1,(Dx2>=Nghx)?Ngx:Nhx+Dx2,
				   (Dy2>=Nghy)?Ngy:Nhy+Dy2,
				   (Dz2>=Nghz)?Ngz:Nhz+Dz2);
	H=Read(PBSh,(Dx2<0)?-Dx2:0,
		    (Dy2<0)?-Dy2:0,
		    (Dz2<0)?-Dz2:0,(Dy1>=Nghy)?Ngy-Dy1:Nhy,
				   (Dy1>=Nghy)?Ngy-Dy1:Nhy,
				   (Dz1>=Nghz)?Ngz-Dz1:Nhz);

	for (dz=Dz1; dz<=Dz2; dz++) {
	    if (dz<0) {
		zg=0; zh=(-dz); Nz=(dz>=Nghz)?Ngz:Nhz+dz;
	    }
	    else {
		zg=dz; zh=0; Nz=(dz>=Nghz)?Ngz-dz:Nhz;
	    }
	    for (dy=Dy1; dy<=Dy2; dy++) {
		if (dy<0) {
		    yg=0; yh=(-dy); Ny=(dy>=Nghy)?Ngy:Nhy+dy;
		}
		else {
		    yg=dy; yh=0; Ny=(dy>=Nghy)?Ngy-dy:Nhy;
		}
		for (dx=Dx1; dx<=Dx2; dx++) {
		    if (dx<0) {
			xg=0; xh=(-dx); Nx=(dx>=Nghx)?Ngx:Nhx+dx;
		    }
		    else {
			xg=dx; xh=0; Nx=(dx>=Nghx)?Ngx-dx:Nhx;
		    }

		    pels=0; sqrs=0.0;
		    for (BSg=PBSg+zg, Gz=G+zg,
			 BSh=PBSh+zh, Hz=H+zh, z=0; z<Nz; z++) {

			Bgh=BSg->base-BSh->base;
			Sg=(BSg++)->step;
			Sh=(BSh++)->step;

			for (Gy=(*(Gz++))+yg,
			     Hy=(*(Hz++))+yh, y=0; y<Ny; y++)
			for (Gx=(*(Gy++))+xg,
			     Hx=(*(Hy++))+xh, x=0; x<Nx; x++) {
			    g=(*(Gx++));
			    h=(*(Hx++));
			    if (g && h) {
				++pels;
				diff=Bgh+Sg*(double)g-Sh*(double)h;
				sqrs+=(diff*diff);
			    }
			}
		    }
		    if (pels)
			(void)printf("%d\t%d\t%d\t%lu\t%lf\n",
				     dx,dy,dz,pels,sqrt(sqrs/(double)pels));
		}
	    }
	}
	return 0;
}
