
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

#ifndef	CSI_BS
#define	CSI_BS
#endif

#include "csi.h"

extern void	Error();

static void	Range(Ng,Nh,D,Og,Oh,N)
int		Ng,Nh,D,*Og,*Oh,*N;
{
	if (D<0) {
	    *Og=0; *Oh=(-D); if ((*N=Nh+D)>Ng) *N=Ng;
	}
	else {
	    *Og=D; *Oh=0; if ((*N=Ng-D)>Nh) *N=Nh;
	}
}

static void	Read(pbs,Ox,Oy,Nx,Ny,data)
PBS		*pbs;
int		Ox,Oy,Nx,Ny;
double		**data;
{
	int	Mx,My,y,x;
        Cell	**cell;

	ReadImageFile(pbs->path,&Mx,&My,NULL,&cell,NULL);

	Nx+=Ox;
	Ny+=Oy;

	for (y=My-1; y>=Ny; y--) free(cell[y]);

	for (y=Ny-1; y>=My; y--)
	    for (x=Nx-1; x>=Ox; x--) data[y-Oy][x-Ox]=pbs->base;

	for (; y>=Oy; y--) {
	    for (x=Nx-1; x>=Mx; x--) data[y-Oy][x-Ox]=pbs->base;

	    for (; x>=Ox; x--)
		data[y-Oy][x-Ox]=pbs->base+pbs->step*(double)cell[y][x];

	    free(cell[y]);
	}
	for (; y>=0; y--) free(cell[y]);

	free(cell);
}

#define LEN	1023

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	PBS	*PBSg,*PBSh;
	int	Ngx,Ngy,Ngz,Nhx,Nhy,Nhz,Dx,Dy,Dz,BPS,
		xg,xh,Nx,yg,yh,Ny,zg,zh,Nz,y,len,z,x;
	double	base,step,roof,**G,**H,gh,min,max;
	Cell	**cell;
	char	format[LEN],slice[LEN],path[LEN],
		*mae="memory allocation error.";

	if (argc!=10 && argc!=12)
	    Error(
	"usage : sub_3d dir1 NBS1 dir2 NBS2 Dx Dy Dz {base step} BPS directory"
	    );

	PBSg=CheckSliceImages(argv[1],argv[2],&Ngx,&Ngy,&Ngz,NULL);
	PBSh=CheckSliceImages(argv[3],argv[4],&Nhx,&Nhy,&Nhz,NULL);

	if ((Dx=atoi(argv[5]))<=(-Nhx) || Dx>=Ngx ||
	    (Dy=atoi(argv[6]))<=(-Nhy) || Dy>=Ngy ||
	    (Dz=atoi(argv[7]))<=(-Nhz) || Dz>=Ngz) Error("bad distance.");

	if ((BPS=atoi(argv[argc-2]))<1 || BPS>8*sizeof(Cell)) Error("bad BPS.");

	if (argc==12) {
	    if ((step=atof(argv[9]))<=0.0) Error("bad step.");

	    base=atof(argv[8]); roof=base+step*(double)((1<<BPS)-1);
        }
	Range(Ngx,Nhx,Dx,&xg,&xh,&Nx);
	Range(Ngy,Nhy,Dy,&yg,&yh,&Ny);
	Range(Ngz,Nhz,Dz,&zg,&zh,&Nz);

	if ((cell=(Cell **)(G=ALLOC(double *,Ny)))==NULL ||
	    (H=ALLOC(double *,Ny))==NULL) Error(mae);

	for (y=0; y<Ny; y++)
	    if ((G[y]=ALLOC(double,Nx))==NULL ||
		(H[y]=ALLOC(double,Nx))==NULL) Error(mae);

	len=1; for (z=Nz-1; z>=10; z/=10) ++len;
	sprintf(format,"%%0%dd",len);

	for (z=0; z<Nz; z++) {
	    Read(PBSg+z+zg,xg,yg,Nx,Ny,G);
	    Read(PBSh+z+zh,xh,yh,Nx,Ny,H);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		gh=G[y][x]-H[y][x];

		if (x==0 && y==0)
		    min=max=gh;
		else
		    if (gh<min) min=gh; else if (gh>max) max=gh;
	    }
	    if (argc==10) {
		if (min==max) fputs("no value range (warning).\n",stderr);

		step=(max-min)/(double)((1<<BPS)-1); base=min; roof=max;
	    }
	    sprintf(slice,format,z);
	    printf("%s\t%lf\t%lf\t%lf\t%18.16lf\n",slice,min,max,base,step);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		gh=G[y][x]-H[y][x];

		cell[y][x]=(gh<=base)?0:
			   (gh>=roof)?(Cell)((1<<BPS)-1):
				      (Cell)((gh-base)/step+0.5);
	    }
	    sprintf(path,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(path,Nx,Ny,BPS,cell,NULL);
        }
        return 0;
}
